#ifndef ALIAST_H
#define ALIAST_H
#include "kwdset.h"

//  aliases for type of problem solved
enum problemtypet {stationary_problem=50, nonlinear_stationary_problem=51, 
		   nonstationary_problem=60, nonlinear_nonstationary_problem=61, 
		   discont_nonstat_problem=62, discont_nonlin_nonstat_problem=63, 
		   growing_np_problem=70, growing_np_problem_nonlin=71,
		   hermes=201};
const enumstr problemtypetstr[] = {{"stationary_problem",50},{"nonlinear_stationary_problem",51},
				   {"nonstationary_problem",60},{"nonlinear_nonstationary_problem",61},
				   {"discont_nonstat_problem",62},{"discont_nonlin_nonstat_problem",63},
				   {"growing_np_problem",70},{"growing_np_problem_nonlin",71},
				   {"hermes",201}};
const kwdset problemtypet_kwdset(sizeof(problemtypetstr)/sizeof(*problemtypetstr),problemtypetstr);


//  aliases for number of transported media/variables
enum transmattert {nomedium=0,onemedium=1,twomediacoup=10,threemediacoup=30,fourmediacoup=40};
const enumstr transmattertstr[] = {{"nomedium",0},{"onemedium",1},{"twomediacoup",10},
				    {"threemediacoup",30},{"fourmediacoup",40}};
const kwdset transmattert_kwdset(sizeof(transmattertstr)/sizeof(*transmattertstr),transmattertstr);


//  aliases for names of transported media/variables
enum mednamest {heat=1,moisture=2,heat_moisture=10,moisture_salt=20,
		moisture_salt_krystal=25,moisture_salt_krystal_heat=30};
const enumstr mednameststr[] = {{"heat",1},{"moisture",2},{"heat_moisture",10},{"moisture_salt",20},
				 {"moisture_salt_krystal",25},{"moisture_salt_krystal_heat",30}};
const kwdset mednamest_kwdset(sizeof(mednameststr)/sizeof(*mednameststr),mednameststr);


//aliases for nonlinear solvers
enum nlsolvertype {newtont=1};
const enumstr nlsolvertypestr[] = {{"newtont",1}};
const kwdset nlsolvertype_kwdset(sizeof(nlsolvertypestr)/sizeof(*nlsolvertypestr), nlsolvertypestr);



enum gravityaccelerationt {gr_yes=1,gr_no=0};
const enumstr gravityaccelerationtstr[]={{"gr_yes",1},{"gr_no",0}};
const kwdset gravityaccelerationt_kwdset(sizeof(gravityaccelerationtstr)/sizeof(*gravityaccelerationtstr),gravityaccelerationtstr);


enum transpsolver {fullnewtont=1,modnewtont=2};
const enumstr transpsolverstr[]={{"fullnewtont",1},{"modnewtont",2}};
const kwdset transpsolver_kwdset(sizeof(transpsolverstr)/sizeof(*transpsolverstr),transpsolverstr);


enum transpresiduumtype {fluxest=1,lrhst=2};
const enumstr transpresiduumtypestr[]={{"fluxest",1},{"lrhst",2}};
const kwdset transpresiduumtype_kwdset(sizeof(transpresiduumtypestr)/sizeof(*transpresiduumtypestr),transpresiduumtypestr);


enum proptypet {boundarycondt=2, crosssect=4,
		eltypet=6, matelt=7, initcondt=11, comcodnumt=12,
		loadedget=13, loadsource=15, boconsurf=21};


//  aliases for elements
enum elemtypet {barlint=200,barlintax=201,barquadt=205,barquadtax=206,trlint=210,trlaxisym=211,
		quadlint=215,quadquadt=216,quadlaxisym=217,quadquadtax=218,
		lineartett=220,linearhext=225,quadratichext=226,gen2del=300,trquadt=500,trqaxisym=501};//trquadt and trqaxisym not  used
const enumstr elemtypetstr[] = {{"barlint",200},{"barlintax",201},{"barquadt",205},{"barquadtax",206},
				{"trlint",210},{"trlaxisym",211},{"trquadt",500},{"trqaxisym",501},
				{"quadlint",215},{"quadquadt",216},{"quadlaxisym",217},{"quadquadtax",218},
				{"lineartett",220},{"linearhext",225},{"quadratichext",226},{"gen2del",300}};
const kwdset elemtypet_kwdset(sizeof(elemtypetstr)/sizeof(*elemtypetstr), elemtypetstr);


//  aliases for material models
enum mattypet {isotransmat=100,cernyconcrete=101,nlisotransmat=102,homomat=110,lincoupledmat=120,
	       bazantpedersen=150,pedersen=151, milly = 154,kunzel=155,grunewald=156,devries=157,kunzel2=158,simplediscmat=159,
	       concreteB=160,baroghelB=161,C60baroghelB=165,C30baroghelB=166,
	       o30bazantB=167,C60bazantB=168,C30bazantB=169,glasgow=170, carb1mat=171,
	       salt1mat=200,salt2mat=201,salt3mat=202,salt4mat=203,radiationmat=251,
	       soilmat1=300,discontisotrmat=401,
	       cementhydrmat=500,sejtkr=600,consolawf2=700};
const enumstr mattypetstr[] = {{"isotransmat",100},{"cernyconcrete",101},{"nlisotransmat",102},{"homomat",110},{"lincoupledmat",120},
			       {"bazantpedersen",150},{"pedersen",151}, {"milly",154}, {"kunzel",155}, {"grunewald",156},
			       {"devries",157}, {"kunzel2",158}, {"simplediscmat",159},
			       {"concreteB",160},{"baroghelB",161},{"C60baroghelB",165},{"C30baroghelB",166},
			       {"o30bazantB",167},{"C60bazantB",168},{"C30bazantB",169},{"glasgow",170},{"carb1mat",171},
			       {"salt1mat",200},{"salt2mat",201},{"salt3mat",202},{"salt4mat",203},
			       {"radiationmat",251},
			       {"soilmat1",300},{"discontisotrmat",401},
			       {"cementhydrmat",500},{"sejtkr",600},{"consolawf2",700}};
const kwdset mattypet_kwdset(sizeof(mattypetstr)/sizeof(*mattypetstr), mattypetstr);


//  aliases for cross sections
enum crsectypet {nocrosssectiont=0,crsec1dt=1,crsec2dt=2,crsec3dt=3};
const enumstr crsectypetstr[] = {{"nocrosssectiont",0},{"crsec1dt",1},{"crsec2dt",2},{"crsec3dt",3}};
const kwdset crsectypet_kwdset(sizeof(crsectypetstr)/sizeof(*crsectypetstr), crsectypetstr);


// aliases for boundary conditions
enum bocontypet {no_bc=0, presc_vals=1, presc_flux=2, det_climcond=3, gen_climcond=4, presc_trmiss=30};
const enumstr bocontypetstr[] = {{"no_bc",0}, {"presc_vals",1}, {"presc_flux",2}, {"det_climcond",3}, {"gen_climcond",4}, {"presc_trmiss",30}};
const kwdset bocontype_kwdset(sizeof(bocontypetstr)/sizeof(*bocontypetstr), bocontypetstr);


//  aliases for position on elements, where values are computed
enum elempositiont {nowheret=0,intptst=1,enodest=2,userdefinedt=3};
const enumstr elempositiontstr[] = {{"nowheret",0},{"intptst",1},{"enodest",2},{"userdefinedt",3}};
const kwdset elempositiont_kwdset(sizeof(elempositiontstr)/sizeof(*elempositiontstr), elempositiontstr);


enum strastret {grad=0,flux=1,othert=2,eqothert=3,pgrad=4,pflux=5};
const enumstr strastretstr[] = {{"grad",0}, {"flux",1}, {"othert",2}, {"eqothert",3}, {"pgrad",4}, {"pflux",5}};
const kwdset strastret_kwdset(sizeof(strastretstr)/sizeof(*strastretstr),strastretstr);

// aliases for format of graphics output
enum graphftt {grftt_no=0, grftt_open_dx=1, grftt_femcad=2, grftt_gid=3, grftt_gid_sep=4,grftt_vtk=5};
const enumstr graphfttstr[] = {{"grftt_no",0}, {"grftt_open_dx",1}, {"grftt_femcad",2}, 
			       {"grftt_gid",3}, {"grftt_gid_sep",4}, {"grftt_vtk",5}};
const kwdset graphftt_kwdset(sizeof(graphfttstr)/sizeof(*graphfttstr), graphfttstr);


// aliases for type of printed unknown
enum prunkt {pr_unknowns=1,pr_gradients=2,pr_fluxes=3,pr_stepidt=6,pr_apploadt=7,pr_othert=8,pr_eqothert=9};
const enumstr prunktstr[] = {{"pr_unknowns",1}, {"pr_gradients",2}, {"pr_fluxes",3}, 
			      {"pr_stepidt",6}, {"pr_apploadt",7}, {"pr_othert",8}, {"pr_eqothert",9}};
const kwdset prunkt_kwdset(sizeof(prunktstr)/sizeof(*prunktstr), prunktstr);


//aliases for time printing
enum timetypeprint {secondst=1,minutest=2,hourst=3,dayst=4};
const enumstr timetypeprintstr[] = {{"secondst",1}, {"minutest",2}, {"hourst",3}, {"dayst",4}};
const kwdset timetypeprint_kwdset(sizeof(timetypeprintstr)/sizeof(*timetypeprintstr), timetypeprintstr);

//  aliases for type of heat source
enum sourcetype {matfunction=1,concrete_heat=2,cement_hydration=5};
const enumstr sourcetypestr[] = {{"matfunction",1}, {"concrete_heat",2}, {"cement_hydration",5}};
const kwdset sourcetype_kwdset(sizeof(sourcetypestr)/sizeof(*sourcetypestr),sourcetypestr);


enum sourceloc {nod=1,elem=2};

//  aliases for type of macro-micro problem correspondence
enum macromicrotype {elem_domain=1,aggregate_domain=2,eff_aggregate_domain=3};
const enumstr macromicrotypestr[] = {{"elem_domain",1}, {"aggregate_domain",2}, {"eff_aggregate_domain",3}};
const kwdset macromicrotype_kwdset(sizeof(macromicrotypestr)/sizeof(*macromicrotypestr),macromicrotypestr);


#endif

