#include "edget.h"
#include "globalt.h"
#include "globmatt.h"
#include <string.h>

edget::edget (void)
{
  
  tm1 = NULL;
  tm2 = NULL;
  idm1 = NULL;
  idm2 = NULL;
  
  jumpfn=NULL;
  jumpln=NULL;
}

edget::~edget (void)
{
  delete [] tm1;
  delete [] tm2;
  delete [] idm1;
  delete [] idm2;
  
  delete [] jumpfn;
  delete [] jumpln;
}

/**
   function initializes edge
   
   @param edid - edge id
   
   JK, 6.8.2008
*/
void edget::init (long edid)
{
  long i,eid1,eid2,ntm;
  
  //  number of the first adjacent element
  eid1=Gtt->gedges[edid].adjel[0];
  //  number of the second adjacent element
  eid2=Gtt->gedges[edid].adjel[1];
  
  //  number of transported media
  ntm=Tp->ntm*Tp->ntm;

  //  allocation of arrays for material types and id of materials
  tm1 = new mattypet [ntm];
  idm1 = new long [ntm];
  tm2 = new mattypet [ntm];
  idm2 = new long [ntm];
  
  for (i=0;i<ntm;i++){
    tm1[i]=Tt->elements[eid1].tm[i];
    idm1[i]=Tt->elements[eid1].idm[i];

    tm2[i]=Tt->elements[eid2].tm[i];
    idm2[i]=Tt->elements[eid2].idm[i];
  }
  
  jumpfn = new double [ntm];
  jumpln = new double [ntm];
}

/**
   function computes jumps across the edge
   
   @param edid - edge id
   
   JK, 6.8.2008
*/
void edget::compute_jump (long edid)
{
  long lcid,ndofn,fn1,fn2,ln1,ln2;
  double *nvf1,*invf1,*nvf2,*invf2,*nvl1,*invl1,*nvl2,*invl2;

  //  load case id must be equal to zero in this type of problem
  lcid=0;
  
  //  first nodes
  fn1 = Gtt->gedges[edid].nlist[0];
  fn2 = Gtt->gedges[edid].nlist[1];
  //  last nodes
  ln1 = Gtt->gedges[edid].nlist[2];
  ln2 = Gtt->gedges[edid].nlist[3];
  
  //  number of DOFs on selected node
  ndofn = Gtt->give_ndofn (fn1);
  nvf1 = new double [ndofn];
  nvf2 = new double [ndofn];
  invf1 = new double [ndofn];
  invf2 = new double [ndofn];
  //  nodal values on selected nodes
  nodval (lcid,nvf1,fn1);
  nodval (lcid,nvf2,fn2);
  //  initial nodal values on selected nodes
  initnodval (lcid,invf1,fn1);
  initnodval (lcid,invf2,fn2);
  
  //  number of DOFs on selected node
  ndofn = Gtt->give_ndofn (ln1);
  nvl1 = new double [ndofn];
  nvl2 = new double [ndofn];
  invl1 = new double [ndofn];
  invl2 = new double [ndofn];
  //  nodal values on selected nodes
  nodval (lcid,nvl1,ln1);
  nodval (lcid,nvl2,ln2);
  //  initial nodal values on selected nodes
  initnodval (lcid,invl1,ln1);
  initnodval (lcid,invl2,ln2);
  
  //  computation of jumps along discontinuity
  Tm->values_transformation (tm1[0],idm1[0],nvf1,invf1,tm2[0],idm2[0],nvf2,invf2,jumpfn);
  Tm->values_transformation (tm1[0],idm1[0],nvl1,invl1,tm2[0],idm2[0],nvl2,invl2,jumpln);
  
  delete [] nvf1;
  delete [] nvf2;
  delete [] nvl1;
  delete [] nvl2;
}


/**
   function computes jumps in selected node on edge
   
   @param edid - edge id
   @param nodid - node id
          nodid=1 - jumps in first node are computed
	  nodid=2 - jumps in last node are computed
   @param lcid - load case id
   @param ncf - normal component of flux
   
   JK, 11.2.2011
*/
void edget::compute_node_jump (long edid, long nodid,long lcid,double ncf)
{
  long ndofn,n1,n2;
  double *nv1,*inv1,*nv2,*inv2;

  //  default values of node numbers
  n1=-1;
  n2=-1;
  
  if (nodid==1){
    //  first nodes
    n1 = Gtt->gedges[edid].nlist[0];
    n2 = Gtt->gedges[edid].nlist[1];
  }
  if (nodid==2){
    //  last nodes
    n1 = Gtt->gedges[edid].nlist[2];
    n2 = Gtt->gedges[edid].nlist[3];
  }
  
  if (n1<0 || n2<0){
    print_err("wrong node is required",__FILE__,__LINE__,__func__);
  }
  
  //  number of DOFs on selected node
  ndofn = Gtt->give_ndofn (n1);
  
  nv1 = new double [ndofn];
  nv2 = new double [ndofn];
  inv1 = new double [ndofn];
  inv2 = new double [ndofn];

  //  nodal values on selected nodes
  nodval (lcid,nv1,n1);
  nodval (lcid,nv2,n2);

  //  initial nodal values on selected nodes
  initnodval (lcid,inv1,n1);
  initnodval (lcid,inv2,n2);
  
  //  computation of jumps along discontinuity in selected node
  if (nodid==1){
    Tm->values_transformation (tm1[0],idm1[0],nv1,inv1,tm2[0],idm2[0],nv2,inv2,jumpfn);
  }
  if (nodid==2){
    Tm->values_transformation (tm1[0],idm1[0],nv1,inv1,tm2[0],idm2[0],nv2,inv2,jumpln);
  }

  delete [] nv1;
  delete [] nv2;
  delete [] inv1;
  delete [] inv2;
}
